LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.all;
USE IEEE.STD_LOGIC_UNSIGNED.all;
USE IEEE.STD_LOGIC_ARITH.all;
USE STD.TEXTIO.all;

ENTITY DataCombining_tb IS 
END DataCombining_tb;

ARCHITECTURE behavioral OF DataCombining_tb IS 

  SIGNAL     Din       : STD_LOGIC_VECTOR(19 downto 0);
  SIGNAL     rden_in   : STD_LOGIC;
  SIGNAL     ADC_range : STD_LOGIC_VECTOR(11 downto 0);
  SIGNAL     clk       : STD_LOGIC;
  SIGNAL     reset     : STD_LOGIC;

  SIGNAL     Dout      : STD_LOGIC_VECTOR(17 downto 0);
  SIGNAL     rden_out  : STD_LOGIC;


COMPONENT DataCombining
    PORT
        (    Din       : in  STD_LOGIC_VECTOR(19 downto 0);
             rden_in   : in  STD_LOGIC;
             ADC_Range : in  STD_LOGIC_VECTOR(11 downto 0);
             clk       : in  STD_LOGIC;
             reset     : in  STD_LOGIC;
    
             Dout      : out STD_LOGIC_VECTOR(17 downto 0);
             rden_out  : out STD_LOGIC
       	);
END COMPONENT;
   
   
BEGIN

  ADC_Range <= X"FFC";
 
DC1: DataCombining
    PORT MAP
        (    Din       => Din,
             rden_in   => rden_in,
             ADC_Range => ADC_Range,
             clk       => clk,
             reset     => reset,
    
             Dout      => Dout,
             rden_out  => rden_out
      	 );
      	 
      	 
PClock: PROCESS
  BEGIN
    clk <= '0';
    WAIT FOR 12.5 ns;

    clk <= '1';
    WAIT FOR 12.5 ns;
  END PROCESS;


PReset: PROCESS
  BEGIN
    reset <= '1';
    WAIT FOR 30 ns;
    
    reset <= '0';
    WAIT;
  END PROCESS;
       

Stimulus: PROCESS
  FILE InFile            : TEXT OPEN READ_MODE IS "stimulus.csv";
  VARIABLE L             : LINE;
  VARIABLE DataIn_int    : INTEGER;
  VARIABLE DataIn        : STD_LOGIC_VECTOR(19 DOWNTO 0);
  VARIABLE ADC_in        : STD_LOGIC_VECTOR(11 DOWNTO 0);
  VARIABLE Cnt_in        : STD_LOGIC_VECTOR(7 DOWNTO 0);  
  VARIABLE ADC_down_cnt  : STD_LOGIC_VECTOR(11 DOWNTO 0) := (OTHERS => '1');
  
  BEGIN
    WAIT FOR 32.5 ns;
    
    WHILE NOT ENDFILE(InFile) LOOP
      READLINE(InFile, L);
      READ(L, DataIn_int);    
	  
      -- No more data to read on current line: transmit data read from file as-is
      IF L'LENGTH = 0 THEN
        Din <= CONV_STD_LOGIC_VECTOR(DataIn_int, 20);
      -- "Standard" transmission
      ELSE
        DataIn := CONV_STD_LOGIC_VECTOR(DataIn_int, 18) & "00";
        ADC_in := DataIn(11 DOWNTO 0);
        Cnt_in := DataIn(19 DOWNTO 12);
      
        IF Cnt_in < X"FF" AND ADC_in > ADC_down_cnt THEN
          Din(19 DOWNTO 12) <= Cnt_in + 1;
        ELSE
          Din(19 DOWNTO 12) <= Cnt_in;
        END IF;
      
        ADC_down_cnt := ADC_down_cnt - ADC_in;
        Din(11 DOWNTO 0) <= ADC_down_cnt;
      END IF;
      
      rden_in <= '1';
      WAIT FOR 20 ns;
      
      rden_in <= '0';
      WAIT FOR 280 ns;
    END LOOP;
        
    FILE_CLOSE(InFile);
    REPORT "End of simulation" SEVERITY FAILURE;
  END PROCESS;
  
END behavioral;