-- Testbench for the RCC block of the BLETC Threshold Comparator firmware
--   with automatic evaluation of results

LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.all;
USE IEEE.STD_LOGIC_UNSIGNED.all;
USE IEEE.STD_LOGIC_ARITH.all;
USE IEEE.STD_LOGIC_TEXTIO.all;
USE STD.TEXTIO.all;

LIBRARY WORK;

ENTITY RCC_tb IS 
END RCC_tb;

ARCHITECTURE behavioral OF RCC_tb IS 

  TYPE TxDataType IS
  RECORD
    LKup    : STD_LOGIC;
    CID     : STD_LOGIC_VECTOR(15 DOWNTO 0);
    St1     : STD_LOGIC_VECTOR(15 DOWNTO 0);
    St2     : STD_LOGIC_VECTOR(15 DOWNTO 0);
    D80_1   : STD_LOGIC_VECTOR(79 DOWNTO 0);
    D80_2   : STD_LOGIC_VECTOR(79 DOWNTO 0);
    DAC1_2  : STD_LOGIC_VECTOR(15 DOWNTO 0);
    DAC3_4  : STD_LOGIC_VECTOR(15 DOWNTO 0);
    DAC5_6  : STD_LOGIC_VECTOR(15 DOWNTO 0);
    DAC7_8  : STD_LOGIC_VECTOR(15 DOWNTO 0);
    FID     : STD_LOGIC_VECTOR(15 DOWNTO 0);
    CRCmask : STD_LOGIC_VECTOR(31 DOWNTO 0);
    CRC     : STD_LOGIC_VECTOR(31 DOWNTO 0);    -- CRC as calculated: transmitted CRC is (CRC XOR CRCmask) !!!
  END RECORD;


  SIGNAL     clk           : STD_LOGIC;
  SIGNAL     clk_TLK_ChA   : STD_LOGIC;
  SIGNAL     clk_TLK_ChB   : STD_LOGIC;
  SIGNAL     TXEN_TXER_ChA : STD_LOGIC_VECTOR(1 downto 0);
  SIGNAL     Data_ChA      : STD_LOGIC_VECTOR(15 downto 0);
  SIGNAL     TXEN_TXER_ChB : STD_LOGIC_VECTOR(1 downto 0);
  SIGNAL     Data_ChB      : STD_LOGIC_VECTOR(15 downto 0);
  SIGNAL     reset         : STD_LOGIC;

  SIGNAL     CheckedSignal : STD_LOGIC_VECTOR(79 downto 0);
  SIGNAL     DReady_PartA  : STD_LOGIC;
  SIGNAL     DReady_PartB  : STD_LOGIC;
  SIGNAL     Dump          : STD_LOGIC;
  SIGNAL     ERRA          : STD_LOGIC;
  SIGNAL     ERRB          : STD_LOGIC;
  SIGNAL     ERRC          : STD_LOGIC;
  SIGNAL     ERRD          : STD_LOGIC;
  SIGNAL     ERRF          : STD_LOGIC;
  SIGNAL     StatusA       : STD_LOGIC_VECTOR(15 downto 0);
  SIGNAL     StatusB       : STD_LOGIC_VECTOR(15 downto 0);
  SIGNAL     SOF_ChA       : STD_LOGIC;
  SIGNAL     SOF_ChB       : STD_LOGIC;
  SIGNAL     CID           : STD_LOGIC_VECTOR(15 downto 0);
  SIGNAL     FID           : STD_LOGIC_VECTOR(15 downto 0);
  SIGNAL     DAC1_2        : STD_LOGIC_VECTOR(15 downto 0);
  SIGNAL     DAC3_4        : STD_LOGIC_VECTOR(15 downto 0);
  SIGNAL     DAC5_6        : STD_LOGIC_VECTOR(15 downto 0);
  SIGNAL     DAC7_8        : STD_LOGIC_VECTOR(15 downto 0);
  SIGNAL     LKdown_ChA    : STD_LOGIC;
  SIGNAL     LKdown_ChB    : STD_LOGIC;
  
  SIGNAL     TxData_ChA    : TxDataType;
  SIGNAL     TxData_ChB    : TxDataType;
  SIGNAL     TxDone        : STD_LOGIC;
  
  SIGNAL     NewFrame      : STD_LOGIC := '0';
  SIGNAL     Done          : STD_LOGIC;

  
COMPONENT RCC
     PORT
         (     clk           :  IN  STD_LOGIC;
               ignoreCFC     :  IN  STD_LOGIC;
               reset         :  IN  STD_LOGIC;
               clk_TLK_ChA   :  IN  STD_LOGIC;
               clk_TLK_ChB   :  IN  STD_LOGIC;
               Status_ChA    :  IN  STD_LOGIC_VECTOR(1 DOWNTO 0);
               Data_ChA      :  IN  STD_LOGIC_VECTOR(15 DOWNTO 0);
               Status_ChB    :  IN  STD_LOGIC_VECTOR(1 DOWNTO 0);
               Data_ChB      :  IN  STD_LOGIC_VECTOR(15 DOWNTO 0);
               
               CheckedSignal :  OUT  STD_LOGIC_VECTOR(79 DOWNTO 0);
               ReEn_Part_A   :  OUT  STD_LOGIC;
               ReEn_Part_B   :  OUT  STD_LOGIC;
               Dump          :  OUT  STD_LOGIC;
               ERRA          :  OUT  STD_LOGIC;
               ERRB          :  OUT  STD_LOGIC;
               ERRC          :  OUT  STD_LOGIC;
               ERRD          :  OUT  STD_LOGIC;
               ERRF          :  OUT  STD_LOGIC;
               StatusA       :  OUT  STD_LOGIC_VECTOR(15 DOWNTO 0);
               StatusB       :  OUT  STD_LOGIC_VECTOR(15 DOWNTO 0);
               SOF_ChA       :  OUT  STD_LOGIC;
               SOF_ChB       :  OUT  STD_LOGIC;
               CID           :  OUT  STD_LOGIC_VECTOR(15 DOWNTO 0);
               FID           :  OUT  STD_LOGIC_VECTOR(15 DOWNTO 0);
               DAC_1_2x      :  OUT  STD_LOGIC_VECTOR(15 DOWNTO 0);
               DAC_3_4x      :  OUT  STD_LOGIC_VECTOR(15 DOWNTO 0);
               DAC_5_6x      :  OUT  STD_LOGIC_VECTOR(15 DOWNTO 0);
               DAC_7_8x      :  OUT  STD_LOGIC_VECTOR(15 DOWNTO 0);
               D16_ChA       :  OUT  STD_LOGIC_VECTOR(15 DOWNTO 0);
               D16_ChB       :  OUT  STD_LOGIC_VECTOR(15 DOWNTO 0);
               Dvalid_ChA    :  OUT  STD_LOGIC;
               Dvalid_ChB    :  OUT  STD_LOGIC;
               ERRTLK_ChA    :  OUT  STD_LOGIC;
               ERRTLK_ChB    :  OUT  STD_LOGIC;
               LKdown_ChA    :  OUT  STD_LOGIC;
               LKdown_ChB    :  OUT  STD_LOGIC
    );
END COMPONENT;


FUNCTION nextCRC 
    (
        Data               :    STD_LOGIC_VECTOR(15 DOWNTO 0);
        CRC                :    STD_LOGIC_VECTOR(31 DOWNTO 0)
    )

    RETURN STD_LOGIC_VECTOR IS
        VARIABLE D        :    STD_LOGIC_VECTOR(15 DOWNTO 0);
        VARIABLE C        :    STD_LOGIC_VECTOR(31 DOWNTO 0);
        VARIABLE NewCRC   :    STD_LOGIC_VECTOR(31 DOWNTO 0);

  BEGIN

    D := Data;
    C := CRC;

    NewCRC(0) := D(12) XOR D(10) XOR D(9) XOR D(6) XOR D(0) XOR C(16) XOR 
                 C(22) XOR C(25) XOR C(26) XOR C(28);
    NewCRC(1) := D(13) XOR D(12) XOR D(11) XOR D(9) XOR D(7) XOR D(6) XOR 
                 D(1) XOR D(0) XOR C(16) XOR C(17) XOR C(22) XOR C(23) XOR 
                 C(25) XOR C(27) XOR C(28) XOR C(29);
    NewCRC(2) := D(14) XOR D(13) XOR D(9) XOR D(8) XOR D(7) XOR D(6) XOR 
                 D(2) XOR D(1) XOR D(0) XOR C(16) XOR C(17) XOR C(18) XOR 
                 C(22) XOR C(23) XOR C(24) XOR C(25) XOR C(29) XOR C(30);
    NewCRC(3) := D(15) XOR D(14) XOR D(10) XOR D(9) XOR D(8) XOR D(7) XOR 
                 D(3) XOR D(2) XOR D(1) XOR C(17) XOR C(18) XOR C(19) XOR 
                 C(23) XOR C(24) XOR C(25) XOR C(26) XOR C(30) XOR C(31);
    NewCRC(4) := D(15) XOR D(12) XOR D(11) XOR D(8) XOR D(6) XOR D(4) XOR 
                 D(3) XOR D(2) XOR D(0) XOR C(16) XOR C(18) XOR C(19) XOR 
                 C(20) XOR C(22) XOR C(24) XOR C(27) XOR C(28) XOR C(31);
    NewCRC(5) := D(13) XOR D(10) XOR D(7) XOR D(6) XOR D(5) XOR D(4) XOR 
                 D(3) XOR D(1) XOR D(0) XOR C(16) XOR C(17) XOR C(19) XOR 
                 C(20) XOR C(21) XOR C(22) XOR C(23) XOR C(26) XOR C(29);
    NewCRC(6) := D(14) XOR D(11) XOR D(8) XOR D(7) XOR D(6) XOR D(5) XOR 
                 D(4) XOR D(2) XOR D(1) XOR C(17) XOR C(18) XOR C(20) XOR 
                 C(21) XOR C(22) XOR C(23) XOR C(24) XOR C(27) XOR C(30);
    NewCRC(7) := D(15) XOR D(10) XOR D(8) XOR D(7) XOR D(5) XOR D(3) XOR 
                 D(2) XOR D(0) XOR C(16) XOR C(18) XOR C(19) XOR C(21) XOR 
                 C(23) XOR C(24) XOR C(26) XOR C(31);
    NewCRC(8) := D(12) XOR D(11) XOR D(10) XOR D(8) XOR D(4) XOR D(3) XOR 
                 D(1) XOR D(0) XOR C(16) XOR C(17) XOR C(19) XOR C(20) XOR 
                 C(24) XOR C(26) XOR C(27) XOR C(28);
    NewCRC(9) := D(13) XOR D(12) XOR D(11) XOR D(9) XOR D(5) XOR D(4) XOR 
                 D(2) XOR D(1) XOR C(17) XOR C(18) XOR C(20) XOR C(21) XOR 
                 C(25) XOR C(27) XOR C(28) XOR C(29);
    NewCRC(10) := D(14) XOR D(13) XOR D(9) XOR D(5) XOR D(3) XOR D(2) XOR 
                  D(0) XOR C(16) XOR C(18) XOR C(19) XOR C(21) XOR C(25) XOR 
                  C(29) XOR C(30);
    NewCRC(11) := D(15) XOR D(14) XOR D(12) XOR D(9) XOR D(4) XOR D(3) XOR 
                  D(1) XOR D(0) XOR C(16) XOR C(17) XOR C(19) XOR C(20) XOR 
                  C(25) XOR C(28) XOR C(30) XOR C(31);
    NewCRC(12) := D(15) XOR D(13) XOR D(12) XOR D(9) XOR D(6) XOR D(5) XOR 
                  D(4) XOR D(2) XOR D(1) XOR D(0) XOR C(16) XOR C(17) XOR 
                  C(18) XOR C(20) XOR C(21) XOR C(22) XOR C(25) XOR C(28) XOR 
                  C(29) XOR C(31);
    NewCRC(13) := D(14) XOR D(13) XOR D(10) XOR D(7) XOR D(6) XOR D(5) XOR 
                  D(3) XOR D(2) XOR D(1) XOR C(17) XOR C(18) XOR C(19) XOR 
                  C(21) XOR C(22) XOR C(23) XOR C(26) XOR C(29) XOR C(30);
    NewCRC(14) := D(15) XOR D(14) XOR D(11) XOR D(8) XOR D(7) XOR D(6) XOR 
                  D(4) XOR D(3) XOR D(2) XOR C(18) XOR C(19) XOR C(20) XOR 
                  C(22) XOR C(23) XOR C(24) XOR C(27) XOR C(30) XOR C(31);
    NewCRC(15) := D(15) XOR D(12) XOR D(9) XOR D(8) XOR D(7) XOR D(5) XOR 
                  D(4) XOR D(3) XOR C(19) XOR C(20) XOR C(21) XOR C(23) XOR 
                  C(24) XOR C(25) XOR C(28) XOR C(31);
    NewCRC(16) := D(13) XOR D(12) XOR D(8) XOR D(5) XOR D(4) XOR D(0) XOR 
                  C(0) XOR C(16) XOR C(20) XOR C(21) XOR C(24) XOR C(28) XOR 
                  C(29);
    NewCRC(17) := D(14) XOR D(13) XOR D(9) XOR D(6) XOR D(5) XOR D(1) XOR 
                  C(1) XOR C(17) XOR C(21) XOR C(22) XOR C(25) XOR C(29) XOR 
                  C(30);
    NewCRC(18) := D(15) XOR D(14) XOR D(10) XOR D(7) XOR D(6) XOR D(2) XOR 
                  C(2) XOR C(18) XOR C(22) XOR C(23) XOR C(26) XOR C(30) XOR 
                  C(31);
    NewCRC(19) := D(15) XOR D(11) XOR D(8) XOR D(7) XOR D(3) XOR C(3) XOR 
                  C(19) XOR C(23) XOR C(24) XOR C(27) XOR C(31);
    NewCRC(20) := D(12) XOR D(9) XOR D(8) XOR D(4) XOR C(4) XOR C(20) XOR 
                  C(24) XOR C(25) XOR C(28);
    NewCRC(21) := D(13) XOR D(10) XOR D(9) XOR D(5) XOR C(5) XOR C(21) XOR 
                  C(25) XOR C(26) XOR C(29);
    NewCRC(22) := D(14) XOR D(12) XOR D(11) XOR D(9) XOR D(0) XOR C(6) XOR 
                  C(16) XOR C(25) XOR C(27) XOR C(28) XOR C(30);
    NewCRC(23) := D(15) XOR D(13) XOR D(9) XOR D(6) XOR D(1) XOR D(0) XOR 
                  C(7) XOR C(16) XOR C(17) XOR C(22) XOR C(25) XOR C(29) XOR 
                  C(31);
    NewCRC(24) := D(14) XOR D(10) XOR D(7) XOR D(2) XOR D(1) XOR C(8) XOR 
                  C(17) XOR C(18) XOR C(23) XOR C(26) XOR C(30);
    NewCRC(25) := D(15) XOR D(11) XOR D(8) XOR D(3) XOR D(2) XOR C(9) XOR 
                  C(18) XOR C(19) XOR C(24) XOR C(27) XOR C(31);
    NewCRC(26) := D(10) XOR D(6) XOR D(4) XOR D(3) XOR D(0) XOR C(10) XOR 
                  C(16) XOR C(19) XOR C(20) XOR C(22) XOR C(26);
    NewCRC(27) := D(11) XOR D(7) XOR D(5) XOR D(4) XOR D(1) XOR C(11) XOR 
                  C(17) XOR C(20) XOR C(21) XOR C(23) XOR C(27);
    NewCRC(28) := D(12) XOR D(8) XOR D(6) XOR D(5) XOR D(2) XOR C(12) XOR 
                  C(18) XOR C(21) XOR C(22) XOR C(24) XOR C(28);
    NewCRC(29) := D(13) XOR D(9) XOR D(7) XOR D(6) XOR D(3) XOR C(13) XOR 
                  C(19) XOR C(22) XOR C(23) XOR C(25) XOR C(29);
    NewCRC(30) := D(14) XOR D(10) XOR D(8) XOR D(7) XOR D(4) XOR C(14) XOR 
                  C(20) XOR C(23) XOR C(24) XOR C(26) XOR C(30);
    NewCRC(31) := D(15) XOR D(11) XOR D(9) XOR D(8) XOR D(5) XOR C(15) XOR 
                  C(21) XOR C(24) XOR C(25) XOR C(27) XOR C(31);

    RETURN NewCRC;

  END nextCRC;
  

-- Data Frame Format
-- ----------------------
-- |         CID        |
-- ----------------------
-- |       Status1      |
-- ----------------------
-- |       Status2      |
-- ----------------------
-- |   Cnt1   |   ADC1  |
-- ----------------------
-- | ADC1 | Cnt2 | ADC2 |
-- ----------------------
-- |   ADC2   |   Cnt3  |
-- ----------------------
-- |     ADC3    | Cnt4 |
-- ----------------------
-- | Cnt4 |    ADC4     |
-- ----------------------
-- |   Cnt5   |   ADC5  |
-- ----------------------
-- | ADC5 | Cnt6 | ADC6 |
-- ----------------------
-- |   ADC6   |   Cnt7  |
-- ----------------------
-- |     ADC8    | Cnt8 |
-- ----------------------
-- | Cnt8 |    ADC8     |
-- ----------------------
-- |   DAC1   |   DAC2  |
-- ----------------------
-- |   DAC3   |   DAC4  |
-- ----------------------
-- |   DAC5   |   DAC6  |
-- ----------------------
-- |   DAC7   |   DAC8  |
-- ----------------------
-- |         FID        |
-- ----------------------
-- |        CRC1        |
-- ----------------------
-- |        CRC2        |
-- ----------------------
PROCEDURE transmitFrame(SIGNAL CLK: IN STD_LOGIC; VARIABLE StimFile: IN TEXT; SIGNAL TXEN_TXER: OUT STD_LOGIC_VECTOR(1 DOWNTO 0); SIGNAL Dout: OUT STD_LOGIC_VECTOR(15 DOWNTO 0); SIGNAL TxData: OUT TxDataType ) IS
  VARIABLE L            : LINE;
  VARIABLE EN           : NATURAL RANGE 0 TO 1;
  VARIABLE Data         : STD_LOGIC_VECTOR(15 DOWNTO 0);
  VARIABLE CRC32        : STD_LOGIC_VECTOR(31 DOWNTO 0) := (OTHERS => '1');
  

  BEGIN

    LOOP
      READLINE(StimFile, L);
      
      -- Skip lines beginning with -: comments
      EXIT WHEN L.all(1) /= '-';
    END LOOP;
    
    -- Check Enable
    READ(L, EN);
    IF EN = 1 THEN
      TxData.LKup <= '1';
    
      -- Carrier Extend <K23.7>; TX_ER = 1, TX_EN = 0
      WAIT UNTIL FALLING_EDGE(CLK);
      TXEN_TXER <= "01";
      Dout <= X"F7F7";

      -- Begin transmission: TX_ER = 0, TX_EN = 1
      WAIT UNTIL FALLING_EDGE(CLK);
      TXEN_TXER <= "10";

      -- CID St1 St2 Cnt1ADC1 ADC1Cnt2ADC2 ADC2Cnt3 ADC3Cnt4 Cnt4ADC4 Cnt5ADC5 ADC5Cnt6ADC6 ADC6Cnt7 ADC7Cnt8 Cnt8ADC8
      -- DAC1DAC2 DAC3DAC4 DAC5DAC6 DAC7DAC8 FID CRC1 CRC2
      FOR I IN 1 TO 18 LOOP     
        -- Hexadecimal Read
          HREAD(L, Data);
          CRC32 := nextCRC(Data, CRC32);
          
          CASE I IS
            WHEN 1    =>  TxData.CID    <= Data;
            WHEN 2    =>  TxData.St1    <= Data;
            WHEN 3    =>  TxData.St2    <= Data;
            WHEN 4    =>  TxData.D80_1(15 DOWNTO  0) <= Data;
            WHEN 5    =>  TxData.D80_1(31 DOWNTO 16) <= Data;
            WHEN 6    =>  TxData.D80_1(47 DOWNTO 32) <= Data;
            WHEN 7    =>  TxData.D80_1(63 DOWNTO 48) <= Data;
            WHEN 8    =>  TxData.D80_1(79 DOWNTO 64) <= Data;
            WHEN 9    =>  TxData.D80_2(15 DOWNTO  0) <= Data;
            WHEN 10   =>  TxData.D80_2(31 DOWNTO 16) <= Data;
            WHEN 11   =>  TxData.D80_2(47 DOWNTO 32) <= Data;
            WHEN 12   =>  TxData.D80_2(63 DOWNTO 48) <= Data;
            WHEN 13   =>  TxData.D80_2(79 DOWNTO 64) <= Data;
            WHEN 14   =>  TxData.DAC1_2 <= Data;
            WHEN 15   =>  TxData.DAC3_4 <= Data;
            WHEN 16   =>  TxData.DAC5_6 <= Data;
            WHEN 17   =>  TxData.DAC7_8 <= Data;
            WHEN 18   =>  TxData.FID    <= Data;
          END CASE;
          
          Dout <= Data;
        WAIT UNTIL FALLING_EDGE(CLK);       
      END LOOP;
     
      -- Simulation of CRC errors: XOR masks
      HREAD(L, Data);       TxData.CRCmask(31 DOWNTO 16) <= Data; TxData.CRC(31 DOWNTO 16) <= CRC32(31 DOWNTO 16);
      Dout <= CRC32(31 DOWNTO 16) XOR Data;
      WAIT UNTIL FALLING_EDGE(CLK);
      HREAD(L, Data);       TxData.CRCmask(15 DOWNTO  0) <= Data; TxData.CRC(15 DOWNTO  0) <= CRC32(15 DOWNTO  0);
      Dout <= CRC32(15 DOWNTO 0) XOR Data;
      WAIT UNTIL FALLING_EDGE(CLK);
    
      -- IDLE: <D5.6> <K28.5>, <D16.2> <K28.5>; TX_ER = 0, TX_EN = 0
      TXEN_TXER <= "00";
      Dout <= X"C5BC";
      WAIT UNTIL FALLING_EDGE(CLK);
      Dout <= X"50BC";

    ELSE
      TxData.LKup <= '0';
    END IF;
    
  END PROCEDURE;


BEGIN
  
RCC1: RCC
     PORT MAP
        (     clk           => clk,
              clk_TLK_ChA   => clk_TLK_ChA,
              clk_TLK_ChB   => clk_TLK_ChB,
              ignoreCFC     => '0',
              reset         => reset,
              Status_ChA    => TXEN_TXER_ChA,
              Data_ChA      => Data_ChA,
              Status_ChB    => TXEN_TXER_ChB,
              Data_ChB      => Data_ChB,
          
              CheckedSignal => CheckedSignal,
              ReEn_Part_A   => DReady_PartA,
              ReEn_Part_B   => DReady_PartB,
              Dump          => Dump,
              ERRA          => ERRA,
              ERRB          => ERRB,
              ERRC          => ERRC,
              ERRD          => ERRD,
              ERRF          => ERRF,
              StatusA       => StatusA,
              StatusB       => StatusB,
              SOF_ChA       => SOF_ChA,
              SOF_ChB       => SOF_ChB,
              CID           => CID,
              FID           => FID,
              DAC_1_2x      => DAC1_2,
              DAC_3_4x      => DAC3_4,
              DAC_5_6x      => DAC5_6,
              DAC_7_8x      => DAC7_8,
              D16_ChA       => OPEN,
              D16_ChB       => OPEN,
              Dvalid_ChA    => OPEN,
              Dvalid_ChB    => OPEN,
              ERRTLK_ChA    => OPEN,
              ERRTLK_ChB    => OPEN,
              LKdown_ChA    => LKdown_ChA,
              LKdown_ChB    => LKdown_ChB
    );
 
 
PClock: PROCESS
  BEGIN
    clk <= '0';
    WAIT FOR 12.5 ns;

    clk <= '1';
    WAIT FOR 12.5 ns;
  END PROCESS;

PReset: PROCESS
  BEGIN
    reset <= '1';
    WAIT FOR 30 ns;
    
    reset <= '0';
    
    WAIT;
  END PROCESS;

PNewFrame: PROCESS
  BEGIN
    WAIT FOR 20037.5 ns;

    WHILE Done /= '1' LOOP
      NewFrame <= '1', '0' AFTER 10 ns;
      WAIT FOR 40000 ns;
    END LOOP;
    
    WAIT FOR 500 ns;    
    REPORT "End of simulation" SEVERITY FAILURE;
  END PROCESS;
 
  clk_TLK_ChA <= TRANSPORT clk AFTER 2.5 ns;
  clk_TLK_ChB <= TRANSPORT clk AFTER 1.5 ns;



Stimulus_ChA: PROCESS
  FILE StimFile_ChA        : TEXT OPEN READ_MODE IS "Stimulus_ChA.txt";
  
  BEGIN
    Data_ChA <= X"50BC";
    TXEN_TXER_ChA <= "00";
    
    WHILE NOT ENDFILE(StimFile_ChA) LOOP
      WAIT UNTIL RISING_EDGE(NewFrame);
      TxDone <= '0';
      
      transmitFrame(clk_TLK_ChA, StimFile_ChA, TXEN_TXER_ChA, Data_ChA, TxData_ChA);

      WAIT FOR 5 ns;
      TxDone <= '1';
    END LOOP;
    
    FILE_CLOSE(StimFile_ChA);
    Done <= '1';

    WAIT;
  END PROCESS;
  
Stimulus_ChB: PROCESS
  FILE StimFile_ChB        : TEXT OPEN READ_MODE IS "Stimulus_ChB.txt";
  
  BEGIN
    Data_ChB <= X"50BC";
    TXEN_TXER_ChB <= "00";
    
    WHILE NOT ENDFILE(StimFile_ChB) LOOP
      WAIT UNTIL RISING_EDGE(NewFrame);
      TxDone <= '0';
      
      transmitFrame(clk_TLK_ChB, StimFile_ChB, TXEN_TXER_ChB, Data_ChB, TxData_ChB);

      WAIT FOR 5 ns;
      TxDone <= '1';
    END LOOP;
    
    FILE_CLOSE(StimFile_ChB);
    Done <= '1';

    WAIT;
  END PROCESS;

Checker: PROCESS
  VARIABLE  TxData    : TxDataType;
  VARIABLE  LkAER_p   : STD_LOGIC := '1';
  VARIABLE  LkBER_p   : STD_LOGIC := '1';
  VARIABLE  FID_p     : STD_LOGIC_VECTOR(15 DOWNTO 0);
  
  -- Local error flag variables
  VARIABLE  ERRA_l    : STD_LOGIC;
  VARIABLE  ERRB_l    : STD_LOGIC;
  VARIABLE  ERRD_l    : STD_LOGIC;
  VARIABLE  ERRF_l    : STD_LOGIC;
  VARIABLE  ERRFpc    : STD_LOGIC;
  VARIABLE  ERRFc     : STD_LOGIC;

  
  BEGIN
    WAIT UNTIL TxDone = '1';

    -- Clear local error flags
    ERRA_l := '0';    
    ERRB_l := '0';    
    ERRD_l := '0';    
    ERRF_l := '0';    
    
    -- Link B is up and has no errors
    IF (TxData_ChB.LKup = '1') AND (TxData_ChB.CRCmask = X"00000000") THEN
      TxData   := TxData_ChB;
    -- Link B is down or has errors; data from Link A is taken as default
    ELSE
      TxData   := TxData_ChA;
    END IF;
    
    -- Both links are down
    IF (TxData_ChA.LKup = '0') AND (TxData_ChB.LKup = '0') THEN
      WAIT UNTIL (ERRA OR ERRB OR ERRC OR ERRD OR ERRF) = '1' FOR 1000 ns;
      IF (ERRA OR ERRB OR ERRC OR ERRD OR ERRF) = '1' THEN
        REPORT "Both links are down, error flag INCORRECTLY RAISED!" SEVERITY ERROR;
        WAIT FOR 30 ns;    -- Wait one clock cycle and some more
      END IF;
      
      IF DUMP = '1' THEN
        REPORT "Both links are down, dump correctly generated";
      ELSE
        REPORT "Both links are down, DUMP MISSING!";
      END IF;
    -- At least one of the links is up
    ELSE
      -- Compare data received to what was transmitted
      WAIT UNTIL DReady_PartA = '1';        -- WAIT UNTIL is edge-triggered
      WAIT FOR 5 ns;
      
      -- Carry out check only if transmission had no CRC errors
      IF (TxData.CRCmask = X"00000000") THEN
        IF CheckedSignal /= TxData.D80_1 THEN
          REPORT "D80.1 reception ERROR!" SEVERITY ERROR;
        END IF;
      END IF;
      
      WAIT UNTIL DReady_PartB = '1';
      WAIT FOR 5 ns;

      -- Carry out check only if transmission had no CRC errors
      IF (TxData.CRCmask = X"00000000") THEN
        IF CheckedSignal /= TxData.D80_2 THEN
          REPORT "D80.2 reception ERROR!" SEVERITY ERROR;
        END IF;
        IF StatusA /= TxData.St1 THEN
          REPORT "StatusA reception ERROR!" SEVERITY ERROR;
        END IF;
        IF StatusB /= TxData.St2 THEN
          REPORT "StatusB reception ERROR!" SEVERITY ERROR;
        END IF;
        IF CID /= TxData.CID THEN
          REPORT "CID reception ERROR!" SEVERITY ERROR;
        END IF;
        IF FID /= TxData.FID THEN
          REPORT "FID reception ERROR!" SEVERITY ERROR;
        END IF;
        IF DAC1_2 /= TxData.DAC1_2 THEN
          REPORT "DAC1_2 reception ERROR!" SEVERITY ERROR;
        END IF;
        IF DAC3_4 /= TxData.DAC3_4 THEN
          REPORT "DAC3_4 reception ERROR!" SEVERITY ERROR;
        END IF;
        IF DAC5_6 /= TxData.DAC5_6 THEN
          REPORT "DAC5_6 reception ERROR!" SEVERITY ERROR;
        END IF;
        IF DAC7_8 /= TxData.DAC7_8 THEN
          REPORT "DAC7_8 reception ERROR!" SEVERITY ERROR;
        END IF;
      END IF;
      
      -- Wait one clock cycle for error flags; CRC checks
      WAIT UNTIL clk = '1';
      WAIT FOR 5 ns;
      IF (TxData_ChA.LKup = '1') AND (TxData_ChA.CRCmask /= X"00000000") THEN       
        ERRA_l := '1';
        
        IF (ERRA = '1') THEN
          REPORT "ChA CRC error correctly reported";
        ELSE
          REPORT "ChA CRC error NOT REPORTED!" SEVERITY ERROR;
        END IF;
      ELSE
        IF (ERRA = '1') THEN
          REPORT "ChA CRC error INCORRECTLY REPORTED!" SEVERITY ERROR;
        END IF;
      END IF;
      
      IF (TxData_ChB.LKup = '1') AND (TxData_ChB.CRCmask /= X"00000000") THEN
        ERRB_l := '1';
        
        IF (ERRB = '1') THEN
          REPORT "ChB CRC error correctly reported";
        ELSE
          REPORT "ChB CRC error NOT REPORTED!" SEVERITY ERROR;
        END IF;
      ELSE
        IF (ERRB = '1') THEN
          REPORT "ChB CRC error INCORRECTLY REPORTED!" SEVERITY ERROR;
        END IF;
      END IF;
      
      IF (TxData_ChA.LKup = '1') AND (TxData_ChB.LKup = '1') AND ((TxData_ChA.CRC XOR TxData_ChA.CRCmask) /= (TxData_ChB.CRC XOR TxData_ChB.CRCmask)) THEN
        IF (ERRC = '1') THEN
          REPORT "CRC comparison error correctly reported";
        ELSE
          REPORT "CRC comparison error NOT REPORTED!" SEVERITY ERROR;
        END IF;
      ELSE
        IF (ERRC = '1') THEN
          REPORT "CRC comparison error INCORRECTLY REPORTED!" SEVERITY ERROR;
        END IF;
      END IF;
      
      -- Both links are healthy: do link-to-link CID comparison
      IF (TxData_ChA.LKup = '1') AND (TxData_ChA.CRCmask = X"00000000") AND (TxData_ChB.LKup = '1') AND (TxData_ChB.CRCmask = X"00000000") AND (TxData_ChA.CID /= TxData_ChB.CID) THEN
        ERRD_l := '1';
      
        IF (ERRD = '1') THEN
          REPORT "CID comparison error correctly reported";
        ELSE
          REPORT "CID comparison error NOT REPORTED!" SEVERITY ERROR;
        END IF;
      ELSE
        IF (ERRD = '1') THEN
          REPORT "CID comparison error INCORRECTLY REPORTED!" SEVERITY ERROR;
        END IF;
      END IF;
        
        
      -- Evaluate ERRFpc and ERRFc -> ERRF
      IF ( (TxData.LKup = '1') AND (TxData.CRCmask = X"00000000") ) AND ( (LkAER_p = '0') OR (LkBER_p = '0') ) AND (TxData.FID /= FID_p + 1) THEN
        ERRFpc := '1';
      ELSE
        ERRFpc := '0';
      END IF;
      
      IF (TxData_ChA.LKup = '1') AND (TxData_ChA.CRCmask = X"00000000") AND (TxData_ChB.LKup = '1') AND (TxData_ChB.CRCmask = X"00000000") AND (TxData_ChA.FID /= TxData_ChB.FID) THEN
        ERRFc := '1';
      ELSE
        ERRFc := '0';
      END IF;                
      
      IF (ERRFpc = '1') OR (ERRFc = '1') THEN
        ERRF_l := '1';
        
        IF (ERRF = '1') THEN
          IF (ERRFpc = '1') THEN
            REPORT "ERRFpc correctly reported";
          END IF;

          IF (ERRFc = '1') THEN
            REPORT "ERRFc correctly reported";
          END IF;
        ELSE
          IF (ERRFpc = '1') THEN
            REPORT "ERRFpc NOT REPORTED!" SEVERITY ERROR;
          END IF;

          IF (ERRFc = '1') THEN
            REPORT "ERRFc NOT REPORTED!" SEVERITY ERROR;
          END IF;
        END IF;
      ELSE
        IF (ERRF = '1') THEN
          REPORT "ERRF INCORRECTLY REPORTED!" SEVERITY ERROR;
        END IF;
      END IF;
    
      -- Wait one clock cycle for dump signal
      WAIT UNTIL clk = '1';
      WAIT FOR 5 ns;      
      IF ( ((ERRA_l = '1') OR (TxData_ChA.LKup = '0')) AND ((ERRB_l = '1') OR (TxData_ChB.LKup = '0')) ) OR (ERRD_l = '1') OR (ERRF_l = '1') THEN
        IF DUMP = '1' THEN
          REPORT "Dump correctly generated by error flags";
        ELSE
          REPORT "Dump NOT GENERATED by error flags!" SEVERITY ERROR;
        END IF;
      ELSIF DUMP = '1' THEN
        REPORT "Dump INCORRECTLY GENERATED!" SEVERITY ERROR;
      END IF;        
    END IF;
    
    -- Save values for previous frame for FID comparison
    IF (TxData_ChA.LKup = '1') AND (TxData_ChA.CRCmask = X"00000000") THEN
      LkAER_p := '0';
    ELSE
      LkAER_p := '1';
    END IF;
  
    IF (TxData_ChB.LKup = '1') AND (TxData_ChB.CRCmask = X"00000000") THEN
      LkBER_p := '0';
    ELSE
      LkBER_p := '1';
    END IF;
    
    FID_p := TxData.FID;
  END PROCESS;
  
END behavioral;